@@ -6,13 +6,14 @@ pass the names of specific files to format instead.
 
 import argparse
 import functools
+import multiprocessing
 import re
 import subprocess
 import sys
 import textwrap
 from operator import attrgetter
 from pathlib import Path
-from typing import FrozenSet, Match, Tuple
+from typing import FrozenSet, Match, Tuple, Union
 
 import autoflake
 import black
@@ -25,22 +26,41 @@ from pybetter.cli import (
     FixParenthesesInReturn,
 )
 
+
+def _fallback(source: str, **kw: object) -> Tuple[str, object]:
+    return source, None  # pragma: no cover
+
+
 try:
-    from teyit import rewrite_source as _teyit_rewrite_source
+    from teyit import refactor_until_deterministic as _teyit_refactor
 except ImportError:  # pragma: no cover  # on Python 3.9
     assert sys.version_info < (3, 9)
+    _teyit_refactor = _fallback
 
-    def _teyit_rewrite_source(source: str) -> Tuple[str, object]:
-        return source, None
+try:
+    from hypothesis.extra.codemods import refactor as _hypothesis_refactor
+except ImportError:  # pragma: no cover  # optional integration
+
+    def _hypothesis_refactor(source_code: str) -> str:
+        return source_code
 
 
-__version__ = "0.2.1"
+# We can't use a try-except here because com2ann does not declare python_requires,
+# and so it is entirely possible to install it on a Python version that it does
+# not support, and nothing goes wrong until you call the function.  We therefore
+# explicitly check the Python version, while waiting on an upstream fix.
+com2ann = _fallback
+if sys.version_info[:2] >= (3, 8):  # pragma: no cover
+    from com2ann import com2ann
+
+
+__version__ = "0.3.0"
 __all__ = ["shed", "docshed"]
 
 _version_map = {
-    black.TargetVersion.PY36: (3, 6),
-    black.TargetVersion.PY37: (3, 7),
-    black.TargetVersion.PY38: (3, 8),
+    k: (int(k.name[2]), int(k.name[3:]))
+    for k in black.TargetVersion
+    if k.value >= black.TargetVersion.PY36.value
 }
 _pybetter_fixers = tuple(
     fix().improve
@@ -76,19 +96,38 @@ def shed(
     if refactor:
         # Some tools assume that the file is multi-line, but empty files are valid input.
         source_code += "\n"
+        # Use com2ann to comvert type comments to annotations on Python 3.8+
+        source_code, _ = com2ann(
+            source_code,
+            drop_ellipsis=True,
+            silent=True,
+            python_minor_version=min_version[1],
+        )
         # Use teyit to replace old unittest.assertX methods on Python 3.9+
-        source_code, _ = _teyit_rewrite_source(source_code)
+        source_code, _ = _teyit_refactor(source_code)
+        # Apply Hypothesis codemods to fix any deprecated code
+        source_code = _hypothesis_refactor(source_code)
         # Then apply pybetter's fixes with libcst
         tree = libcst.parse_module(source_code)
         for fixer in _pybetter_fixers:
-            tree = fixer(tree)
+            try:
+                # Might raise e.g. https://github.com/Instagram/LibCST/issues/446
+                newtree = fixer(tree)
+                # Catches e.g. https://github.com/lensvol/pybetter/issues/60
+                compile(newtree.code, "<string>", "exec")
+            except Exception:
+                pass
+            else:
+                tree = newtree
         source_code = tree.code
     # Then shed.docshed (below) formats any code blocks in documentation
     source_code = docshed(source=source_code, first_party_imports=first_party_imports)
     # And pyupgrade - see pyupgrade._fix_file - is our last stable fixer
-    source_code = pyupgrade._fix_tokens(source_code, min_version=min_version)
+    # Calculate separate minver because pyupgrade doesn't have py39-specific logic yet
+    pyupgrade_min_ver = min(min_version, max(pyupgrade.IMPORT_REMOVALS.keys()))
+    source_code = pyupgrade._fix_tokens(source_code, min_version=pyupgrade_min_ver)
     source_code = pyupgrade._fix_percent_format(source_code)
-    source_code = pyupgrade._fix_py3_plus(source_code, min_version=min_version)
+    source_code = pyupgrade._fix_py3_plus(source_code, min_version=pyupgrade_min_ver)
     source_code = pyupgrade._fix_py36_plus(source_code)
 
     # One tricky thing: running `isort` or `autoflake` can "unlock" further fixes
@@ -168,6 +207,7 @@ def docshed(
     return rst_pattern.sub(_rst_match, markdown_pattern.sub(_md_match, source))
 
 
+@functools.lru_cache()
 def _guess_first_party_modules(cwd: str = None) -> FrozenSet[str]:
     """Guess the name of the current package for first-party imports."""
     try:
@@ -185,10 +225,38 @@ def _guess_first_party_modules(cwd: str = None) -> FrozenSet[str]:
     return frozenset(p for p in {Path(base).name} | provides if p.isidentifier())
 
 
+@functools.lru_cache(maxsize=None)
+def _should_format(fname: str) -> bool:
+    return fname.endswith((".md", ".rst")) or autoflake.is_python_file(fname)
+
+
+def _rewrite_on_disk(
+    fname: str, **kwargs: Union[bool, FrozenSet[str]]
+) -> Union[bool, str]:
+    """Return either bool(rewrote the file), or an error message string."""
+    try:
+        with open(fname) as handle:
+            on_disk = handle.read()
+    except (OSError, UnicodeError) as err:
+        # Permissions or encoding issue, or file deleted since last commit.
+        return f"skipping {fname!r} due to {err}"
+    writer = docshed if fname.endswith((".md", ".rst")) else shed
+    try:
+        result = writer(on_disk, **kwargs)
+    except Exception as err:  # pragma: no cover  # bugs are unknown xor fixed ;-)
+        return (
+            f"Internal error formatting {fname!r}: {err}\n"
+            "    Please report this to https://github.com/Zac-HD/shed/issues"
+        )
+    if result != on_disk:
+        with open(fname, mode="w") as fh:
+            fh.write(result)
+    return result != on_disk
+
+
 def cli() -> None:  # pragma: no cover  # mutates things in-place, will test later.
     """Execute the `shed` CLI."""
     # TODO: make this provide useful CLI help and usage hints
-    # TODO: single-file mode with optional min_version support
     parser = argparse.ArgumentParser(prog="shed", description=__doc__.strip())
     parser.add_argument(
         "--refactor",
@@ -205,10 +273,6 @@ def cli() -> None:  # pragma: no cover  # mutates things in-place, will test lat
 
     if args.files:
         all_filenames = args.files
-        for f in all_filenames:
-            if not autoflake.is_python_file(f):
-                print(f"{f!r} does not seem to be a Python file")  # noqa
-                sys.exit(1)
     else:
         # Get all tracked files from `git ls-files`
         try:
@@ -222,27 +286,23 @@ def cli() -> None:  # pragma: no cover  # mutates things in-place, will test lat
         except subprocess.SubprocessError:
             print("Doesn't seem to be a git repo; pass filenames to format.")  # noqa
             sys.exit(1)
+        all_filenames = [f for f in all_filenames if _should_format(f)]
 
-        all_filenames = [
-            f
-            for f in all_filenames
-            if f.endswith((".md", ".rst")) or autoflake.is_python_file(f)
-        ]
+    rewrite = functools.partial(
+        _rewrite_on_disk,
+        first_party_imports=_guess_first_party_modules(),
+        refactor=args.refactor,
+    )
 
-    first_party_imports = _guess_first_party_modules()
-    for fname in all_filenames:
-        try:
-            with open(fname) as handle:
-                on_disk = handle.read()
-        except (OSError, UnicodeError) as err:
-            # Permissions or encoding issue, or file deleted since last commit.
-            print(f"skipping {fname!r} due to {err}")  # noqa
-            continue
-        kwargs = {"refactor": args.refactor, "first_party_imports": first_party_imports}
-        if fname.endswith((".md", ".rst")):
-            result = docshed(on_disk, **kwargs)
-        else:
-            result = shed(on_disk, **kwargs)
-        if result != on_disk:
-            with open(fname, mode="w") as fh:
-                fh.write(result)
+    if len(all_filenames) <= 4:
+        # If we're only formatting a few files, starting up a process pool
+        # probably takes up more time that it saves.
+        for fname in all_filenames:
+            error_msg = rewrite(fname)
+            if isinstance(error_msg, str):
+                print(error_msg)  # noqa
+    else:
+        with multiprocessing.Pool() as pool:
+            for error_msg in pool.imap_unordered(rewrite, all_filenames):
+                if isinstance(error_msg, str):
+                    print(error_msg)  # noqa
